//#define _SHMEM_

//
// node info on device side
//
static int_gpu   ipeo_d;
static int_gpu    igx_d;
static int_gpu   ieo_d[2];
static int4_gpu  ipsite_d;

const unsigned int maxThreads = 256;

unsigned int nextPow2( unsigned int x ) {
    --x;
    x |= x >> 1;
    x |= x >> 2;
    x |= x >> 4;
    x |= x >> 8;
    x |= x >> 16;
    return ++x;
}

///////////////////////////////////////////////////////////////////////////////////
// Kernel routines
///////////////////////////////////////////////////////////////////////////////////
static __global__ void 
new_q_kernel(cu_wqf_eo_blk *y, const int *ipeo, const int *ieo, const int *igx, const int4 *ipsite){
  (*y).igx    = (*igx) + 1;  // count for igx = 1..NGX
  (*y).ieo    = (*ieo);
  (*y).ipeo   = (*ipeo);
  (*y).ipsite.x = (*ipsite).x;
  (*y).ipsite.y = (*ipsite).y;
  (*y).ipsite.z = (*ipsite).z;
  (*y).ipsite.w = (*ipsite).w;
}

static __global__ void 
new_u_kernel(cu_gvf_eo_blk *u, const int *ipeo, const int *ieo, const int *igx, const int4 *ipsite){
  (*u).igx    = (*igx) + 1;  // count for igx = 1..NGX
  (*u).ieo    = (*ieo);
  (*u).ipeo   = (*ipeo);
  (*u).ipsite.x = (*ipsite).x;
  (*u).ipsite.y = (*ipsite).y;
  (*u).ipsite.z = (*ipsite).z;
  (*u).ipsite.w = (*ipsite).w;
}

static __global__ void 
new_clv_kernel(cu_clvf_eo_blk *fclv, const int *ipeo, const int *ieo, const int *igx, const int4 *ipsite){
  (*fclv).igx    = (*igx) + 1;  // count for igx = 1..NGX
  (*fclv).ieo    = (*ieo);
  (*fclv).ipeo   = (*ipeo);
  (*fclv).ipsite.x = (*ipsite).x;
  (*fclv).ipsite.y = (*ipsite).y;
  (*fclv).ipsite.z = (*ipsite).z;
  (*fclv).ipsite.w = (*ipsite).w;
}

static __global__ void 
clear_kernel(const int4 *blk_indx, cu_wqf_eo_blk *y){
  unsigned int ithd = threadIdx.x;
  unsigned int iblk = blk_indx[blockIdx.x].w;
#pragma unroll 6
  for(unsigned int i=0;i<CSH;i++) {
    (*y).blk[iblk].thd[i].f[ithd].x = 0.0f;
    (*y).blk[iblk].thd[i].f[ithd].y = 0.0f;
    (*y).blk[iblk].thd[i].f[ithd].z = 0.0f;
    (*y).blk[iblk].thd[i].f[ithd].w = 0.0f;
  }
}

static __global__ void 
assign_sub_kernel(const int4 *blk_indx, cu_wqf_eo_blk *v, const cu_wqf_eo_blk *y, const cu_wqf_eo_blk *w) {
//
// v = y - w;
//
  unsigned int ithd = threadIdx.x;
  unsigned int iblk = blk_indx[blockIdx.x].w;
  (*v).igx  = (*y).igx;
  (*v).ieo  = (*y).ieo;
  (*v).ipeo = (*y).ipeo;
  (*v).ipsite.x = (*y).ipsite.x;
  (*v).ipsite.y = (*y).ipsite.y;
  (*v).ipsite.z = (*y).ipsite.z;
  (*v).ipsite.w = (*y).ipsite.w;
#pragma unroll 6
  for(unsigned int i=0;i<CSH;i++) {
    float4 vri;
    float4 yri = (*y).blk[iblk].thd[i].f[ithd];
    float4 wri = (*w).blk[iblk].thd[i].f[ithd];
    vri.x = yri.x - wri.x;
    vri.y = yri.y - wri.y;
    vri.z = yri.z - wri.z;
    vri.w = yri.w - wri.w;
    (*v).blk[iblk].thd[i].f[ithd] = vri;
  }
}

static __global__ void 
accum_rmult_add_kernel(const int4 *blk_indx, cu_wqf_eo_blk *w, const cu_wqf_eo_blk *y) { 
//
// w = w * rcoef + y;
//
// rcoef is fetched via texture fetching.
//
  __align__(16) float4 yri,wri;
  unsigned int ithd = threadIdx.x;
  unsigned int iblk = blk_indx[blockIdx.x].w;
  float tcoef = tex1Dfetch(tex_Float,0);

#pragma unroll 6
  for(unsigned int i=0;i<CSH;i++) {
    yri = (*y).blk[iblk].thd[i].f[ithd];
    wri = (*w).blk[iblk].thd[i].f[ithd];
    yri.x += wri.x * (tcoef);
    yri.y += wri.y * (tcoef);
    yri.z += wri.z * (tcoef);
    yri.w += wri.w * (tcoef);
    (*w).blk[iblk].thd[i].f[ithd] = yri;
  }
}

static __global__ void 
accum_nrmult_add_kernel(const int4 *blk_indx, cu_wqf_eo_blk *w, const cu_wqf_eo_blk *y) { 
//
// w = - w * rcoef + y;
//
// rcoef is fetched via texture fetching.
//
  __align__(16) float4 yri,wri;
  unsigned int ithd = threadIdx.x;
  unsigned int iblk = blk_indx[blockIdx.x].w;
  float tcoef = - tex1Dfetch(tex_Float,0);

#pragma unroll 6
  for(unsigned int i=0;i<CSH;i++) {
    yri = (*y).blk[iblk].thd[i].f[ithd];
    wri = (*w).blk[iblk].thd[i].f[ithd];
    yri.x += wri.x * (tcoef);
    yri.y += wri.y * (tcoef);
    yri.z += wri.z * (tcoef);
    yri.w += wri.w * (tcoef);
    (*w).blk[iblk].thd[i].f[ithd] = yri;
  }
}

static __global__ void 
accum_cmult_add_kernel(const int4 *blk_indx, cu_wqf_eo_blk *w, const cu_wqf_eo_blk *y) {
//
// w = w * ccoef + y;
//
// ccoef is fetched via texture fetching.
//
  __align__(16) float cr,ci,cj;
  __align__(16) float4 wr,wi,yr,yi;
  unsigned int ithd = threadIdx.x;
  unsigned int iblk = blk_indx[blockIdx.x].w;
  cuComplex ccoef = tex1Dfetch(tex_Complex,0);

  cr = ccoef.x;
  ci = ccoef.y;
  cj =-ccoef.y;

#pragma unroll 3
  for(unsigned int i=0;i<COL;i++) {
    yr = (*y).blk[iblk].thd[i    ].f[ithd];
    yi = (*y).blk[iblk].thd[i+COL].f[ithd];
    wr = (*w).blk[iblk].thd[i    ].f[ithd];
    wi = (*w).blk[iblk].thd[i+COL].f[ithd];

    yr.x += cr * wr.x;
    yr.x += cj * wi.x;
    yi.x += cr * wi.x;
    yi.x += ci * wr.x;

    yr.y += cr * wr.y;
    yr.y += cj * wi.y;
    yi.y += cr * wi.y;
    yi.y += ci * wr.y;

    yr.z += cr * wr.z;
    yr.z += cj * wi.z;
    yi.z += cr * wi.z;
    yi.z += ci * wr.z;

    yr.w += cr * wr.w;
    yr.w += cj * wi.w;
    yi.w += cr * wi.w;
    yi.w += ci * wr.w;
    (*w).blk[iblk].thd[i    ].f[ithd] = yr;
    (*w).blk[iblk].thd[i+COL].f[ithd] = yi;
  }
}

static __global__ void 
accum_add_mult_kernel(const int4 *blk_indx, cu_wqf_eo_blk *w, const cu_wqf_eo_blk *y){
//
// w = w + ccoef * y;
//
  __align__(16) float cr,ci,cj;
  __align__(16) float4 wr,wi,yr,yi;
  unsigned int ithd = threadIdx.x;
  unsigned int iblk = blk_indx[blockIdx.x].w;

  cuComplex ccoef = tex1Dfetch(tex_Complex,0);

  cr = ccoef.x;
  ci = ccoef.y;
  cj =-ccoef.y;

#pragma unroll 3
  for(unsigned int i=0;i<COL;i++) {
    yr = (*y).blk[iblk].thd[i    ].f[ithd];
    yi = (*y).blk[iblk].thd[i+COL].f[ithd];
    wr = (*w).blk[iblk].thd[i    ].f[ithd];
    wi = (*w).blk[iblk].thd[i+COL].f[ithd];

    wr.x += cr * yr.x;
    wr.x += cj * yi.x;
    wi.x += cr * yi.x;
    wi.x += ci * yr.x;

    wr.y += cr * yr.y;
    wr.y += cj * yi.y;
    wi.y += cr * yi.y;
    wi.y += ci * yr.y;

    wr.z += cr * yr.z;
    wr.z += cj * yi.z;
    wi.z += cr * yi.z;
    wi.z += ci * yr.z;

    wr.w += cr * yr.w;
    wr.w += cj * yi.w;
    wi.w += cr * yi.w;
    wi.w += ci * yr.w;

    (*w).blk[iblk].thd[i    ].f[ithd] = wr;
    (*w).blk[iblk].thd[i+COL].f[ithd] = wi;
  }

}

////////////////////////
// Reduction Routines
////////////////////////
static __global__ void 
prod_kernel(const int4 *blk_indx, const cu_wqf_eo_blk *y, const cu_wqf_eo_blk *w, cuComplex gri[]){
//
// inner product in each block.
//
//    y : input quark field 
//    w : input quark field 
//  gri[block] :  = <y|w>[block]
//
  unsigned int ithd = threadIdx.x;
  unsigned int iblk = blk_indx[blockIdx.x].w;

  float ywr = 0.0f;
  float yxr = 0.0f;
  float ywi = 0.0f;
  float yxi = 0.0f;

  {
    __align__(16) float4 wr,wi,yr,yi;
#pragma unroll 3
    for(unsigned int i=0;i<COL;i++) {
      yr = (*y).blk[iblk].thd[i    ].f[ithd];
      yi = (*y).blk[iblk].thd[i+COL].f[ithd];
      wr = (*w).blk[iblk].thd[i    ].f[ithd];
      wi = (*w).blk[iblk].thd[i+COL].f[ithd];

      ywr += yr.x * wr.x;
      yxr += yi.x * wi.x;
      ywi += yr.x * wi.x;
      yxi += yi.x * wr.x;

      ywr += yr.y * wr.y;
      yxr += yi.y * wi.y;
      ywi += yr.y * wi.y;
      yxi += yi.y * wr.y;

      ywr += yr.z * wr.z;
      yxr += yi.z * wi.z;
      ywi += yr.z * wi.z;
      yxi += yi.z * wr.z;

      ywr += yr.w * wr.w;
      yxr += yi.w * wi.w;
      ywi += yr.w * wi.w;
      yxi += yi.w * wr.w;
    }
  }

  ywr = ywr + yxr;
  ywi = ywi - yxi;

#ifdef _SF
  {
//
//  Set SF boundary zero (Fortran: itt=1 (C: itt=0))
//
  unsigned int iblk_x = blk_indx[blockIdx.x].x;
  unsigned int iblk_y = blk_indx[blockIdx.x].y;
  unsigned int iblk_z = blk_indx[blockIdx.x].z;
  int ipeo   = (*y).ipeo;
  int igx    = (*y).igx;
  int ieo    = (*y).ieo;
  int iz     = thd_z + iblk_z*_NTHZ;
  int iy     =         iblk_y;
  int ix     =         iblk_x + (igx-1)*_NBX;
  int ieoxyz = (ipeo+ieo+(iz+1)+(iy+1)+(ix+1)+256)%2;
  int it     = (thd_t+1-ieoxyz)*2+ieoxyz-1;
  float flag = (float)(1-(_NTT-it)/_NTT);
  ywr = ywr * flag;
  ywi = ywi * flag;
  }
#endif

  {
    __shared__ __align__(16) float sywr[_CEIL],sywi[_CEIL];
#if _NTHREADS != _CEIL
    sywr[_CEIL - _NTHREADS + threadIdx.x] = 0.0f;
    sywi[_CEIL - _NTHREADS + threadIdx.x] = 0.0f;
#endif
    __syncthreads();

    sywr[threadIdx.x] = ywr;
    sywi[threadIdx.x] = ywi;

    __syncthreads();

    for (unsigned int s = _CEIL/2; s > 32 ; s >>= 1){
      if (threadIdx.x < s) {
        sywr[threadIdx.x] += sywr[threadIdx.x^s];
        sywi[threadIdx.x] += sywi[threadIdx.x^s];
      }
      __syncthreads();
    }
    if (threadIdx.x < 32) {
      sywr[threadIdx.x] += sywr[threadIdx.x^32];
      sywi[threadIdx.x] += sywi[threadIdx.x^32];
      sywr[threadIdx.x] += sywr[threadIdx.x^16];
      sywi[threadIdx.x] += sywi[threadIdx.x^16];
      sywr[threadIdx.x] += sywr[threadIdx.x^8 ];
      sywi[threadIdx.x] += sywi[threadIdx.x^8 ];
      sywr[threadIdx.x] += sywr[threadIdx.x^4 ];
      sywi[threadIdx.x] += sywi[threadIdx.x^4 ];
      sywr[threadIdx.x] += sywr[threadIdx.x^2 ];
      sywi[threadIdx.x] += sywi[threadIdx.x^2 ];
      sywr[threadIdx.x] += sywr[threadIdx.x^1 ];
      sywi[threadIdx.x] += sywi[threadIdx.x^1 ];
    }
    if (threadIdx.x==0) {
       gri[blockIdx.x].x = sywr[0];
       gri[blockIdx.x].y = sywi[0];
    }
  }
}

static __global__ void 
abs2_kernel(const int4 *blk_indx, const cu_wqf_eo_blk *y, float g[]){
//
// squared absolute value in each block
//
//  g[block] = |y|^2[block]
//
  unsigned int ithd = threadIdx.x;
  unsigned int iblk = blk_indx[blockIdx.x].w;

  float ywr = 0.0f;
  {
    __align__(16) float4 yri;
#pragma unroll 6
    for(unsigned int i=0;i<CSH;i++) {
      yri = (*y).blk[iblk].thd[i].f[ithd];
      ywr += yri.x * yri.x;
      ywr += yri.y * yri.y;
      ywr += yri.z * yri.z;
      ywr += yri.w * yri.w;
    }
  }

#ifdef _SF
  {
//
//  Set SF boundary zero (Fortran: itt=1 (C: itt=0))
//
  unsigned int iblk_x = blk_indx[blockIdx.x].x;
  unsigned int iblk_y = blk_indx[blockIdx.x].y;
  unsigned int iblk_z = blk_indx[blockIdx.x].z;
  unsigned int ipeo   = (*y).ipeo;
  unsigned int igx    = (*y).igx;
  unsigned int ieo    = (*y).ieo;
  unsigned int iz     = thd_z + iblk_z*_NTHZ;
  unsigned int iy     =         iblk_y;
  unsigned int ix     =         iblk_x + (igx-1)*_NBX;
  unsigned int ieoxyz = (ipeo+ieo+(iz+1)+(iy+1)+(ix+1)+256)%2;
  unsigned int it     = (thd_t+1-ieoxyz)*2+ieoxyz-1;
  float flag = (float)(1-(_NTT-it)/_NTT);
  ywr = ywr * flag;
  }
#endif

  {
    __shared__ __align__(16) float a[_CEIL];
#if _NTHREADS != _CEIL
    a[_CEIL - _NTHREADS + threadIdx.x] = 0.0f;
#endif

    __syncthreads();

    a[threadIdx.x] = ywr;

    __syncthreads();

    for (unsigned int s = _CEIL/2; s > 32 ; s >>= 1){
      if (threadIdx.x < s) 
        a[threadIdx.x] += a[threadIdx.x^s];
      __syncthreads();
    }

    if (threadIdx.x < 32) {
      a[threadIdx.x] += a[threadIdx.x^32];
      a[threadIdx.x] += a[threadIdx.x^16];
      a[threadIdx.x] += a[threadIdx.x^8 ];
      a[threadIdx.x] += a[threadIdx.x^4 ];
      a[threadIdx.x] += a[threadIdx.x^2 ];
      a[threadIdx.x] += a[threadIdx.x^1 ];
    }
    if (threadIdx.x==0) g[blockIdx.x] = a[0];
  }
}

static __global__ void
reduce_float_kernel(float g[]){
  unsigned int ith = threadIdx.x;
  unsigned int i   = blockIdx.x*(blockDim.x*2) + threadIdx.x;
  extern __shared__ __align__(16) float a[];

  a[ith] = g[i] + g[i+blockDim.x];

  __syncthreads();

  for (unsigned int s = blockDim.x/2; s > 0; s >>= 1){
    if (ith < s) a[ith] += a[ith^s];
    __syncthreads();
  }
  if (ith==0) g[blockIdx.x] = a[0];
}

static __global__ void
reduce_complex_kernel(cuComplex gri[]){
  unsigned int ith = threadIdx.x;
  unsigned int i   = blockIdx.x*(blockDim.x*2) + threadIdx.x;
  extern __shared__ __align__(16) cuComplex ari[];

  ari[ith].x = gri[i].x + gri[i+blockDim.x].x;
  ari[ith].y = gri[i].y + gri[i+blockDim.x].y;

  __syncthreads();

  for (unsigned int s = blockDim.x/2; s > 0; s >>= 1){
    if (ith < s) {
      ari[ith].x += ari[ith^s].x;
      ari[ith].y += ari[ith^s].y;
    }
    __syncthreads();
  }
  if (ith==0) gri[blockIdx.x] = ari[0];
}
////////////////////////////
// End Reduction Routines
////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
// End Kernel routines
///////////////////////////////////////////////////////////////////////////////////

template <const int block, const int thread>
static float
abs2(const int4 *blk_indx, const cu_wqf_eo_blk *y){
//
//  a(block) = |y|^2(block)
//
  float  a;
  float *a_g;
  CUDA_SAFE_CALL(cudaMalloc((void**)&a_g,sizeof(float)*block));
  CUT_CHECK_ERROR("cudaMalloc in abs2 failed");

  abs2_kernel<<<block,thread>>>(blk_indx,y,a_g);
  CUT_CHECK_ERROR("abs2_kernel execution failed");

// chek
//{
//float a_h[block];
//CUDA_SAFE_CALL(cudaMemcpy(a_h,a_g,sizeof(float)*block,cudaMemcpyDeviceToHost));
//CUT_CHECK_ERROR("cudaMemcpy in abs2 failed");
//a = 0.0f;
//for (unsigned int i=0;i < block; i++) {
//  printf("psum[%d]=%16.7e\n",i,a_h[i]);
//  a += a_h[i];
//}
//printf("chk_abs2CPU:%16.7e\n",a);
//}

#if _NTHREADS == _CEIL
//  reduce_float_kernel<<<1,block/2,sizeof(float)*block/2>>>(a_g);
//  CUT_CHECK_ERROR("reduce_float_kernel execution failed");
//  CUDA_SAFE_CALL(cudaMemcpy(&a,a_g,sizeof(float),cudaMemcpyDeviceToHost));
//  CUT_CHECK_ERROR("cudaMemcpy reduce_float_kernel failed");
//  printf("chk_abs2GPU:%16.7e\n",a);

  //
  // GPU side reduction
  //
  {
    unsigned int ndat = block;
    unsigned int threads = (ndat < maxThreads*2) ? nextPow2((ndat + 1)/ 2) : maxThreads;
    unsigned int blocks  = (ndat + threads*2 - 1) / (threads*2);
    dim3 dimGrid(blocks,1,1);
    dim3 dimBlock(threads,1,1);
    size_t size = threads*sizeof(float);

    reduce_float_kernel<<<dimGrid,dimBlock,size>>>(a_g);
    CUT_CHECK_ERROR("reduce_float Kernel execution failed");

#if CUDA_VERSION >= 4000
    CUDA_SAFE_CALL(cudaDeviceSynchronize());
#else
    CUDA_SAFE_CALL(cudaThreadSynchronize());
#endif

    float *hhdat = (float *)malloc(blocks*sizeof(float));
    CUDA_SAFE_CALL(cudaMemcpy(hhdat,a_g,blocks*sizeof(float),cudaMemcpyDeviceToHost));
    a = 0.0f;
    for (unsigned int i=0;i < blocks; i++) {
      a += hhdat[i];
    }
    free(hhdat);
  }
#else
  //
  // Host side reduction
  //
  {
    float a_h[block];
    CUDA_SAFE_CALL(cudaMemcpy(a_h,a_g,sizeof(float)*block,cudaMemcpyDeviceToHost));
    CUT_CHECK_ERROR("cudaMemcpy in abs2 failed");
    a = 0.0f;
    for (unsigned int i=0;i < block; i++) {
      a += a_h[i];
    }
  }
#endif
  CUDA_SAFE_CALL(cudaFree(a_g));
  CUT_CHECK_ERROR("cudaFree in abs2 failed");
  return a;
}

static float
abs2(const cu_wqf_eo *y){
//
// a = |y|^2
//
  float a = 0.0f;
#pragma omp parallel for reduction(+:a)
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    float a_g = abs2<_NBLOCKS,_NTHREADS>(phys_blk_d.g[igx], (*y).g[igx]);
    a += a_g;
  }
#ifndef _singlePU
  sumcast(&a);
#endif

//f_wqf_eo w;
//w.ieo  = (*y).ieo;
//w.ipeo = (*y).ipeo;
//#pragma omp parallel for
//  for (int igx=0;igx < _NGX; igx++) {
//    CUDA_SAFE_CALL(cudaMemcpy(&(w.g[igx]), (*y).g[igx], sizeof(cu_wqf_eo_blk), cudaMemcpyDeviceToHost));
//    CUT_CHECK_ERROR("assign Host<=Device execution failed");
//  }
//float psum = 0.0f;
//for (int igx=1; igx <= _NGX; igx++) {
//for (int iblk=0; iblk< _NBLOCKS; iblk++) {
//for (int ithd=0; ithd< _NTHREADS; ithd++) {
//int idx = phys_blk[iblk].w;
//float4 yy[CSH];
//for (int i=0;i<CSH;i++) {
//  yy[i] = w.g[igx-1].blk[idx].thd[i].f[ithd];
//  psum = psum + yy[i].x * yy[i].x;
//  psum = psum + yy[i].y * yy[i].y;
//  psum = psum + yy[i].z * yy[i].z;
//  psum = psum + yy[i].w * yy[i].w;
//}
//for (int i=0;i<COL;i++) {
//printf("%16.7e %16.7e\n",yy[i].x,yy[i+COL].x);
//printf("%16.7e %16.7e\n",yy[i].y,yy[i+COL].y);
//printf("%16.7e %16.7e\n",yy[i].z,yy[i+COL].z);
//printf("%16.7e %16.7e\n",yy[i].w,yy[i+COL].w);
//}
//}}}
//printf("abs2_a=%16.7e\n",a);
//printf("abs2_p=%16.7e\n",psum);
//if ( a < 0.0f ) {
//  printf("negativ norm is detected.\n");
//  abort();
//}

  return a;
}

template <const int block, const int thread>
static cuComplex
prod(const int4 *blk_indx, const cu_wqf_eo_blk *y, const cu_wqf_eo_blk *w){
//
//  p(block) = <y|w>(block)
//
  cuComplex p;
  cuComplex *p_g;
  CUDA_SAFE_CALL(cudaMalloc((void**)&p_g,sizeof(cuComplex)*block));
  CUT_CHECK_ERROR("cudaMalloc in prod failed");

  prod_kernel<<<block,thread>>>(blk_indx, y, w, p_g);
  CUT_CHECK_ERROR("prod_kernel execution failed");

#if _NTHREADS == _CEIL
  //
  // GPU side reduction
  //
  {
//    reduce_complex_kernel<<<1,block/2,sizeof(cuComplex)*block/2>>>(p_g);
//    CUT_CHECK_ERROR("reduce_compelx_kernel execution failed");
//    CUDA_SAFE_CALL(cudaMemcpy(&p, p_g, sizeof(cuComplex),cudaMemcpyDeviceToHost));
//    CUT_CHECK_ERROR("cudaMemcpy reduce_complex_kernel failed");

    unsigned int ndat = block;
    unsigned int threads = (ndat < maxThreads*2) ? nextPow2((ndat + 1)/ 2) : maxThreads;
    unsigned int blocks  = (ndat + threads*2 - 1) / (threads*2);
    dim3 dimGrid(blocks,1,1);
    dim3 dimBlock(threads,1,1);
    size_t size = threads*sizeof(cuComplex);

    reduce_complex_kernel<<<dimGrid,dimBlock,size>>>(p_g);
    CUT_CHECK_ERROR("reduce_complex Kernel execution failed");

#if CUDA_VERSION >= 4000
    CUDA_SAFE_CALL(cudaDeviceSynchronize());
#else
    CUDA_SAFE_CALL(cudaThreadSynchronize());
#endif

    cuComplex *hhri = (cuComplex *)malloc(blocks*sizeof(cuComplex));
    CUDA_SAFE_CALL(cudaMemcpy(hhri,p_g,blocks*sizeof(cuComplex),cudaMemcpyDeviceToHost));
    p.x = 0.0f;
    p.y = 0.0f;
    for (unsigned int i=0;i < blocks; i++) {
      p.x += hhri[i].x;
      p.y += hhri[i].y;
    }
    free(hhri);
  }
#else
  //
  // Host side reducion
  //
  {
    cuComplex p_h[block];
    CUDA_SAFE_CALL(cudaMemcpy(p_h,p_g,sizeof(cuComplex)*block,cudaMemcpyDeviceToHost));
    CUT_CHECK_ERROR("cudaMemcpy in prod failed");

    p.x = 0.0f;
    p.y = 0.0f;
    for (unsigned int i=0;i < block; i++) {
      p.x += p_h[i].x;
      p.y += p_h[i].y;
    }
  }
#endif
  CUDA_SAFE_CALL(cudaFree(p_g));
  CUT_CHECK_ERROR("cudaFree in prod failed");
  return p;
}

static cuComplex
prod(const cu_wqf_eo *y, const cu_wqf_eo *w){
//
// p = <y|w>
//
  cuComplex p;
  float pr = 0.0f;
  float pi = 0.0f;
#pragma omp parallel for reduction(+:pr,pi)
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    cuComplex p_g = prod<_NBLOCKS,_NTHREADS>(phys_blk_d.g[igx], (*y).g[igx], (*w).g[igx]);
    pr += pr + p_g.x;
    pi += pi + p_g.y;
  }
  p.x = pr;
  p.y = pi;
#ifndef _singlePU
  sumcast(&p);
#endif
  return p;
}

static void 
assign( cu_wqf_eo_blk *w, const cu_wqf_eo_blk *y){
//
// w = y
//
  CUDA_SAFE_CALL(cudaMemcpy(w, y, sizeof(cu_wqf_eo_blk), cudaMemcpyDeviceToDevice));
  CUT_CHECK_ERROR("assign Device<=Device in block execution failed");
}

static void 
assign( cu_wqf_eo *w, const cu_wqf_eo *y){
//
// w = y
//
  (*w).ieo  = (*y).ieo;
  (*w).ipeo = (*y).ipeo;
#pragma omp parallel for
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    CUDA_SAFE_CALL(cudaMemcpy((*w).g[igx], (*y).g[igx], sizeof(cu_wqf_eo_blk), cudaMemcpyDeviceToDevice));
    CUT_CHECK_ERROR("assign Device<=Device execution failed");
  }
}

static void 
assign( f_wqf_eo *w, const cu_wqf_eo *y){
//
// w = y
//
  (*w).ieo  = (*y).ieo;
  (*w).ipeo = (*y).ipeo;
#pragma omp parallel for
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    CUDA_SAFE_CALL(cudaMemcpy(&((*w).g[igx]), (*y).g[igx], sizeof(cu_wqf_eo_blk), cudaMemcpyDeviceToHost));
    CUT_CHECK_ERROR("assign Host<=Device execution failed");
  }
}

static void 
assign( cu_wqf_eo *w, const f_wqf_eo *y){
//
// w = y
//
  (*w).ieo  = (*y).ieo;
  (*w).ipeo = (*y).ipeo;
#pragma omp parallel for
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    CUDA_SAFE_CALL(cudaMemcpy((*w).g[igx], &((*y).g[igx]), sizeof(cu_wqf_eo_blk), cudaMemcpyHostToDevice));
    CUT_CHECK_ERROR("assign Device<=Host execution failed");
  }
}


template <const int block, const int thread>
static void 
assign_sub(const int4 *blk_indx, cu_wqf_eo_blk *v, const cu_wqf_eo_blk *y, const cu_wqf_eo_blk *w){
//
// v = y - w
//
  assign_sub_kernel<<<block,thread>>>(blk_indx, v, y, w);
  CUT_CHECK_ERROR("assign_sub_kernel execution failed");
}

static void 
assign_sub(cu_wqf_eo *v, const cu_wqf_eo *y, const cu_wqf_eo *w){
//
// v = y - w
//
  (*v).ieo  = (*y).ieo;
  (*v).ipeo = (*y).ipeo;
#pragma omp parallel for
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    assign_sub<_NBLOCKS,_NTHREADS>(phys_blk_d.g[igx], (*v).g[igx], (*y).g[igx], (*w).g[igx]);
  }
}

template <const int block, const int thread>
static void 
accum_add_mult(const int4 *blk_indx, cu_wqf_eo_blk *w, const cuComplex *ccoef, const cu_wqf_eo_blk *y){
//
// w = w + ccoef * y
//
  CUDA_SAFE_CALL(cudaBindTexture(0,tex_Complex,ccoef,sizeof(cuComplex)));
  accum_add_mult_kernel<<<block,thread>>>(blk_indx, w, y);
  CUT_CHECK_ERROR("accum_add_mult_kernel execution failed");
  CUDA_SAFE_CALL(cudaUnbindTexture(tex_Complex));
}

static void 
accum_add_mult(cu_wqf_eo *w, const cuComplex_gpu &ccoef, const cu_wqf_eo *y){
//
// w = w + ccoef * y
//
  assert((*w).ieo == (*y).ieo);
#pragma omp parallel for
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    accum_add_mult<_NBLOCKS,_NTHREADS>(phys_blk_d.g[igx], (*w).g[igx], ccoef.g[igx], (*y).g[igx]);
  }
}

template <const int block, const int thread>
static void 
accum_mult_add(const int4 *blk_indx, cu_wqf_eo_blk *w, const float *rcoef, const cu_wqf_eo_blk *y){
//
// w = w * rcoef + y
//
  CUDA_SAFE_CALL(cudaBindTexture(0,tex_Float,rcoef,sizeof(float)));
  accum_rmult_add_kernel<<<block,thread>>>(blk_indx, w, y);
  CUT_CHECK_ERROR("accum_rmult_add_kernel execution failed");
  CUDA_SAFE_CALL(cudaUnbindTexture(tex_Float));
}

static void 
accum_mult_add(cu_wqf_eo *w, const float_gpu &rcoef, const cu_wqf_eo *y){
//
// w = w * rcoef + y
//
  assert((*w).ieo == (*y).ieo);
#pragma omp parallel for
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    accum_mult_add<_NBLOCKS,_NTHREADS>(phys_blk_d.g[igx], (*w).g[igx], rcoef.g[igx], (*y).g[igx]);
  }
}

template <const int block, const int thread>
static void 
accum_mult_add(const int4 *blk_indx, cu_wqf_eo_blk *w, const cuComplex *ccoef, const cu_wqf_eo_blk *y){
//
// w = w * ccoef + y
//
  CUDA_SAFE_CALL(cudaBindTexture(0,tex_Complex,ccoef,sizeof(cuComplex)));
  accum_cmult_add_kernel<<<block,thread>>>(blk_indx, w, y);
  CUT_CHECK_ERROR("accum_cmult_add_kernel execution failed");
  CUDA_SAFE_CALL(cudaUnbindTexture(tex_Complex));
}

static void 
accum_mult_add(cu_wqf_eo *w, const cuComplex_gpu &ccoef, const cu_wqf_eo *y){
//
// w = w * ccoef + y
//
  assert((*w).ieo == (*y).ieo);
#pragma omp parallel for
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    accum_mult_add<_NBLOCKS,_NTHREADS>(phys_blk_d.g[igx], (*w).g[igx], ccoef.g[igx], (*y).g[igx]);
  }
}

////////////////////////////////
static void 
new_q(cu_wqf_eo_blk *y, const int *ipeo, const int *ieo, const int *igx, const int4 *ipsite){
  new_q_kernel<<<1,1>>>(y,ipeo,ieo,igx,ipsite);
  CUT_CHECK_ERROR("new_q_kernel execution failed");
}

static void 
new_q(cu_wqf_eo *y, const int &ieo) {
  (*y).ipeo = _ipeo_;
  (*y).ieo = ieo;
#pragma omp parallel for
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    new_q((*y).g[igx], ipeo_d.g[igx], ieo_d[ieo].g[igx], igx_d.g[igx], ipsite_d.g[igx]);
  }
}

////////////////////////////////
static void 
new_u(cu_gvf_eo_blk *u, const int *ipeo, const int *ieo, const int *igx, const int4 *ipsite){
  new_u_kernel<<<1,1>>>(u,ipeo,ieo,igx,ipsite);
  CUT_CHECK_ERROR("new_u_kernel execution failed");
}

static void 
new_u(cu_gvf_eo *u, const int &ieo){
  (*u).ipeo = _ipeo_;
  (*u).ieo = ieo;
#pragma omp parallel for
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    new_u((*u).g[igx], ipeo_d.g[igx], ieo_d[ieo].g[igx], igx_d.g[igx], ipsite_d.g[igx]);
  }
}

////////////////////////////////
static void 
new_clv(cu_clvf_eo_blk *fclv, const int *ipeo, const int *ieo, const int *igx, const int4 *ipsite){
  new_clv_kernel<<<1,1>>>(fclv,ipeo,ieo,igx,ipsite);
  CUT_CHECK_ERROR("new_clv_kernel execution failed");
}

static void 
new_clv(cu_clvf_eo *fclv, const int &ieo){
  (*fclv).ipeo = _ipeo_;
  (*fclv).ieo = ieo;
#pragma omp parallel for
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    new_clv((*fclv).g[igx], ipeo_d.g[igx], ieo_d[ieo].g[igx], igx_d.g[igx], ipsite_d.g[igx]);
  }
}

////////////////////////////////
static void
clear(cu_wqf_eo *y, const int &ieo){
  (*y).ipeo = _ipeo_;
  (*y).ieo = ieo;
#pragma omp parallel for
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    CUDA_SAFE_CALL(cudaMemset((*y).g[igx], 0, sizeof(cu_wqf_eo_blk)));
    new_q((*y).g[igx], ipeo_d.g[igx], ieo_d[ieo].g[igx], igx_d.g[igx], ipsite_d.g[igx]);
  }
}

////
static void
print_attr(const char *s, const cu_wqf_eo *ye){
  int4 ippss;
  int  iggxx;
  int  ieeoo;
#pragma omp parallel for
  for (int igx=0;igx < _NGX; igx++) {
    switch_gpu_(&igx);
    CUDA_SAFE_CALL(cudaMemcpy(&ippss,&((*ye).g[igx]->ipsite),sizeof(int4),cudaMemcpyDeviceToHost));
    CUDA_SAFE_CALL(cudaMemcpy(&ieeoo,&((*ye).g[igx]->ieo   ),sizeof(int),cudaMemcpyDeviceToHost));
    CUDA_SAFE_CALL(cudaMemcpy(&iggxx,&((*ye).g[igx]->igx   ),sizeof(int),cudaMemcpyDeviceToHost));
    printf("print_attr: %s[igx=%d] ipsite=%d %d %d %d, igx=%d, ieo=%d\n",s,igx,ippss.x,ippss.y,ippss.z,ippss.w,iggxx,ieeoo);
  }
}

