#include <config.h>

module hmc_qcd_class
!
!= HMC algorithm with QCD action
!
  use lattice_class
  use error_class
  use timer_class
  use counter_class
  use hmc_logfile_class
  use hmc_gluon_class
  use hmc_quark_class
  use hmc_status_class
  use onfly_measure_class
  implicit none
  private
  public :: qcd_hmc
  public :: wloops_obj
  public :: new,delete
  public :: print
  public :: print_statistics
  public :: read
  public :: initialize_config
  public :: save,save_cont_config
  public :: save_config
  public :: update_p
  public :: update_q
  public :: hamil
  public :: metropolis
  public :: measurement
  public :: set_current_md_depth
  public :: rev_check

  !
  ! Lattice World
  !
  type(lattice_world),  save :: m_lattice

  !
  ! for reversivility check
  !
  type(vfield_gluon_wg),  save, pointer :: m_u0,m_u1
  type(vfield_gluon_wog), save, pointer :: m_p0,m_p1

  type qcd_hmc
  !
  ! QCD HMC program parameter
  !
    private
    type(vfield_gluon_wg),  pointer ::  u => NULL()  ! current  gauge field
    type(vfield_gluon_wg),  pointer :: u0 => NULL()  ! previous gauge field
    type(vfield_gluon_wog), pointer ::  p => NULL()  ! gauge momentum
    type(vfield_gluon_wog), pointer ::  F => NULL()  ! force field
    type(hmc_gluon),        pointer :: gluon    => NULL()
    type(hmc_quark),        pointer :: quark(:) => NULL()
    real(DP) :: H0  ! initial Hamiltonian
    real(DP) :: H1  !   final Hamiltonian
    real(DP) :: H2  ! for reversivility check
    integer :: number_of_quark        = 0
    integer :: switch_dynamical_quark = SW_OFF
    integer :: switch_measurement     = SW_OFF
    integer :: switch_stout           = SW_OFF
    integer :: idepth = 0                       ! current MD depth
    type(hmc_logfile)  :: force_log
    type(hmc_logfile)  ::  plaq_log
    character(CHARLEN) :: force_log_fname="force."
    character(CHARLEN) ::  plaq_log_fname="plaqhist."
    character(CHARLEN) :: input_file_comment(4)
    integer(8) :: mult0,mult1
    type(wloops_obj) :: WLP0,WLP1,WLP2
    type(hmc_status), pointer :: hmc_status
    character(CHARLEN) :: measurement_fname                    ! onfly measurement input parameter file name
    type(onfly_measurement), pointer :: observables => NULL()  ! observables for onfly measurement
  end type

  type hamiltonian
  !
  ! Contains Action values
  !
    real(DP) :: H   ! total Hamiltonian
    real(DP) :: SU  ! Gauge Action
    real(DP) :: SP  ! Gauge Kinetic tarm
    real(DP) :: SQ  ! Quark Action
  end type

  interface new
  !
  ! Initialize/create HMC QCD action
  !
    module procedure new_qcd_hmc
  end interface

  interface delete
  !
  ! delete HMC QCD action
  !
    module procedure delete_qcd_hmc
  end interface

  interface read
  !
  ! read HMC QCD action parameters
  !
    module procedure read_qcd_hmc
  end interface

  interface print
  !
  ! print out HMC QCD action parameters
  !
    module procedure print_qcd_hmc
  end interface

  interface save
  !
  ! save HMC QCD action parameters
  !
    module procedure save_qcd_hmc
  end interface

  interface print_statistics
  !
  ! print out HMC QCD action job statistics
  !
    module procedure print_stat_qcd_hmc
  end interface

  interface read_cont_config
  !
  ! read field configuration of HMC QCD action for job continuation
  !
    module procedure read_contg_qcd_hmc
  end interface

  interface save_cont_config
  !
  ! save field configuration of HMC QCD action for job continuation
  !
    module procedure save_contg_qcd_hmc
  end interface

  interface save_config
  !
  ! save field configuration of HMC QCD action for measurement
  !
    module procedure save_config_qcd_hmc
  end interface

  interface force
  !
  ! return QCD action force for HMC algorithm
  !
    module procedure force_qcd_hmc
  end interface

  interface update_p
  !
  ! update momentum with QCD action for HMC algorithm
  !
    module procedure update_p_qcd_hmc
  end interface

  interface update_q
  !
  ! update corrdinate with QCD action for HMC algorithm
  !
    module procedure update_q_qcd_hmc
  end interface

  interface hamil
  !
  ! compute total Hmiltonian value for HMC algorithm
  !
    module procedure hamil_qcd_hmc
  end interface

  interface metropolis
  !
  ! do Metropolis test with QCD action for HMC algorithm
  !
    module procedure metropolis_all_qcd_hmc
  end interface

  interface measurement
  !
  ! do on the fly measurment with QCD action for HMC algorithm
  !
    module procedure measurement_onfly_qcd_hmc
  end interface

  interface initialize_config
  !
  ! initalize/load field configuration
  !
    module procedure initialize_config_qcd_hmc
  end interface

  interface set_current_md_depth
  !
  ! set depth id for recuresive MD integrator 
  !
    module procedure set_current_md_depth_qh
  end interface

  interface rev_check
  !
  ! do reversibility check for MD integrator
  !
    module procedure rev_check_qcd_hmc
  end interface

contains

subroutine set_current_md_depth_qh(this,idepth)
  implicit none
  type(qcd_hmc), intent(inout) :: this
  integer, intent(in) :: idepth
  this%idepth = idepth
  return
end subroutine

subroutine new_qcd_hmc(this,status)
!
! Initialize and Create HMC QCD action
!
  use comlib
  implicit none
  type(qcd_hmc), intent(inout) :: this
  type(hmc_status), intent(in), target :: status

  call new(m_lattice)      ! create global lattice

  allocate(this%u,this%u0)
  call new(this%u)
  call new(this%u0)

  this%hmc_status => status

  call new(this%force_log,TRIM(ADJUSTL(this%force_log_fname)))
  call new( this%plaq_log, TRIM(ADJUSTL(this%plaq_log_fname)))

  this%observables => NULL()

  return
end subroutine

subroutine delete_qcd_hmc(this)
!
! Delelete HMC QCD action
!
  implicit none
  type(qcd_hmc), intent(inout) :: this

  call delete(this%gluon)
  if (SW_ON == this%switch_dynamical_quark) then
    call delete(this%quark)
    deallocate(this%quark)
  endif
  call delete(this%plaq_log)
  call delete(this%force_log)

  if ( ASSOCIATED(this%observables) ) then
    call delete(this%observables)
    deallocate(this%observables)
    this%observables => NULL()
  endif

  return
end subroutine

subroutine print_qcd_hmc(this)
!
! Print out HMC QCD parameters on display
!
  implicit none
  type(qcd_hmc), intent(inout) :: this

  if (nodeid == 0) then
    write(*,'("**** Lattice ****")')
  endif
  call print(m_lattice) ! print global lattice

  if (nodeid == 0) then
    write(*,*)
    write(*,'("**** Gluon Action ****")')
  endif
  call print(this%gluon)

  select case (this%switch_dynamical_quark)
  case (SW_OFF)
    if (nodeid == 0) then
      write(*,*)
      write(*,'("**** Quenched QCD ****")')
    endif
  case (SW_ON)
    if (nodeid == 0) then
      write(*,*)
      write(*,'("**** Full QCD ****")')
    endif
    call print(this%quark)
  end select

  select case (this%switch_measurement)
  case (SW_OFF)
  case (SW_ON)
    if (nodeid == 0) then
      write(*,*)
      write(*,'("**** DO measurement onfly ****")')
      write(*,'(" Onfly Measurement Inputfile : ",A)')TRIM(this%measurement_fname)
    endif
  end select
  
  return
end subroutine

subroutine print_stat_qcd_hmc(this,elapse_time)
!
! Print out HMC QCD job statistics on display
!
  use comlib
  implicit none
  type(qcd_hmc), intent(inout) :: this
  real(DP), intent(in) :: elapse_time
  real(DP) :: mult_traj
  real(DP) :: copy_gluon_time
  real(DP) :: copy_quark_time

  mult_traj = REAL(get_count(mult_iter),kind=DP)/get_sweep_count(this%hmc_status)
#ifndef _singlePU
  call comlib_sumcast(mult_traj)
  mult_traj=mult_traj/NPU
#endif

  copy_gluon_time = get_elapse(copy_sfg_time)
  copy_quark_time = get_elapse(copy_fq_time)

  call print_copy_sfg_statistics()

  if (nodeid == 0) then
    write(*,'("                copy_u time (sec) :",F12.4," (",F8.4," %)")')  &
 &        copy_gluon_time, 100*copy_gluon_time/elapse_time
  endif

  call print_copy_fq_statistics()

  if (nodeid == 0) then
    write(*,'("             copy_y(c) time (sec) :",F12.4," (",F8.4," %)")')  &
 &        copy_quark_time, 100*copy_quark_time/elapse_time
    write(*,'("           # of Mult (par 1sweep) :",F12.4)') mult_traj
  endif

  if (SW_ON == this%switch_dynamical_quark) then
    if (nodeid == 0) then
      write(*,'(80("-"))')
      write(*,'("  Quark Statistics")')
    endif
    call print_statistics(this%quark)
  endif

  if (nodeid == 0) then
    write(*,'(80("-"))')
  endif

  return
end subroutine

subroutine read_qcd_hmc(this,iout)
!
! Read HMC QCD job parameters form file indicated by unit id 
!
  use comlib
  use tokenize
  use print_status_class
  implicit none
  type(qcd_hmc), intent(inout) :: this
  integer,       intent(in)    :: iout
  integer :: iqk,nqk
  integer :: nstep,ioerr
  type(tokenizer)  :: token
  real(DP) :: alpha
  integer :: token_len
  character(len=CHARLEN) :: str,part

  call set_tokenizer(token,' ',' ',' ')

  allocate(this%gluon)
  call new(this%gluon,id=0)

  !----------------------
  ! Stout parameters
  !----------------------
  alpha = 0.0_DP
  nstep = 0
  ioerr = 0
  if (nodeid==0) then
    read(iout,'(A)')this%gluon%file_name

    this%switch_stout = SW_OFF
    read(iout,'(A)')str
    part = first_token(token,str,token_len)
    read(part,*)this%number_of_quark
    part = next_token(token,str,token_len)
    if (token_len > 0) then
      this%switch_stout = SW_ON
      read(part,*)alpha
    endif
    part = next_token(token,str,token_len)
    if (token_len > 0) then
      this%switch_stout = SW_ON
      read(part,*)nstep
    endif

  endif

#ifndef _singlePU
  call comlib_bcast(this%gluon%file_name,0)
  call comlib_bcast(this%number_of_quark,0)
  call comlib_bcast(alpha,0)
  call comlib_bcast(nstep,0)
  call comlib_bcast(this%switch_stout,0)
#endif

#ifndef _STOUT
  if (SW_ON == this%switch_stout) then
    call warning_message("This program is not configured with Stout smearing.")
  endif
#endif

  nqk=this%number_of_quark
  allocate(this%quark(1:nqk))
#ifdef _STOUT
  call new(this%quark,alpha,nstep)
#else
  call new(this%quark)
#endif

  if (nodeid==0) then
    do iqk=1,nqk
      read(iout,'(A)')this%quark(iqk)%file_name
    enddo
    read(iout,*)this%switch_dynamical_quark
    read(iout,*)this%switch_measurement

    if ( SW_ON == this%switch_measurement ) then
      read(iout,'(A)')this%measurement_fname
    endif
  endif

#ifndef _singlePU
  do iqk=1,nqk
    call comlib_bcast(this%quark(iqk)%file_name,0)
  enddo
  call comlib_bcast(this%switch_dynamical_quark,0)
  call comlib_bcast(this%switch_measurement,0)
  if ( SW_ON == this%switch_measurement ) then
    call comlib_bcast(this%measurement_fname,0)
  endif
#endif

  !==================================
  ! read gauge parameter
  !==================================
  call read(this%gluon)

  if (SW_ON == this%switch_dynamical_quark) then
    !==================================
    ! read quark parameter
    !==================================
    call read(this%quark)
  endif

  if (nodeid == 0) write(*,'(80("-"))')

  return
end subroutine

subroutine save_qcd_hmc(this,iout)
!
! Save HMC QCD job parameter in a file with unit id iout.
!
  implicit none
  type(qcd_hmc), intent(inout) :: this
  integer,       intent(in)    :: iout
  real(DP) :: alpha
  integer :: nstep
  integer :: iqk,nqk

  if (nodeid == 0) then
    write(iout,'(a)') TRIM(this%gluon%file_name)
    if (SW_ON == this%switch_stout) then
#ifdef _STOUT
      call get_quarks_stout_params(this%quark,alpha,nstep)
      write(iout,'(I3,E15.6,I4)')this%number_of_quark,alpha,nstep
#else
      write(iout,'(I3)')this%number_of_quark
#endif
    else
      write(iout,'(I3)')this%number_of_quark
    endif
    nqk=this%number_of_quark
    do iqk=1,nqk
      write(iout,'(a)')TRIM(this%quark(iqk)%file_name)
    enddo
    write(iout,'(I3)')this%switch_dynamical_quark
    write(iout,'(I3)')this%switch_measurement
    if ( SW_ON == this%switch_measurement ) then
      write(iout,'(A)')TRIM(this%measurement_fname)
    endif
  endif

  return
end subroutine


subroutine read_contg_qcd_hmc(this,fname)
!
! Read gauge configuration for continuation from file
!
  use comlib
  use file_tools_class
  use print_status_class
  use random_class
  implicit none
  type(qcd_hmc),      intent(inout) :: this
  character(CHARLEN), intent(in)    :: fname

  character(CHARLEN) :: job_id
  real(DP) :: plq0,plq1
  integer :: iout

  iout = search_free_file_unit()

  open(iout,file=TRIM(fname),form='unformatted')
  read(iout)this%u
  call read(g_rand,iout)
  read(iout)plq0
  read(iout)job_id
  close(iout)

  call plaquette(this%u,plq1)

  if (abs(plq1-plq0) > 10*EPSILON(plq1)) then
    if (nodeid == 0) then
      write(*,'(" Cont. Gauge PLQ chack sum is inconsistent!")')
      write(*,'("  Old :",E24.16)')plq0
      write(*,'("  New :",E24.16)')plq1
      write(*,'(" Diff :",E24.16)')(plq1-plq0)/plq0
    endif
    call error_stop("")
  endif

  if (nodeid == 0) then
    write(*,'(80("-"))')
    write(*,'("                 JOB ID :",a)')TRIM(job_id)
    write(*,'("               PLQ(OLD) :",F24.16)')plq0
    write(*,'("               PLQ(NEW) :",F24.16)')plq1
    write(*,*)
  endif

  return
end subroutine


subroutine save_contg_qcd_hmc(this,iout)
!
! Save configuration for continuation to a file indicated by an unit id.
!
  use comlib
  use print_status_class
  use random_class
  implicit none
  type(qcd_hmc), intent(inout) :: this
  integer,       intent(in)    :: iout

  real(DP) :: plq


  call plaquette(this%u,plq)

  write(iout)this%u
  call save(g_rand,iout)
  write(iout)plq

  if (nodeid == 0) then
    write(*,'(80("-"))')
    write(*,'(" PLQ :",F24.16)')plq
  endif

  return
end subroutine


subroutine save_config_qcd_hmc(this,iout)
!
! Save configuration for measurment to a file indicated by an unit id.
!
  use hmc_identity
  use print_status_class
  use comlib
  implicit none
  type(qcd_hmc), intent(inout) :: this
  integer,       intent(in)    :: iout
  real(DP) :: plq
  real(DP) :: alpha
  integer  :: nstep

  call plaquette(this%u,plq)

!**************************************
!* write configuration
!**************************************
  write(iout) this%u
!**************************************
!* configuration information
!**************************************
  write(iout) NTX,NTY,NTZ,NTT
  write(iout) NDIMX,NDIMY,NDIMZ
  write(iout) plq
  call save_config(this%gluon,iout)
  if (SW_ON == this%switch_dynamical_quark) then
    if (SW_ON == this%switch_stout) then
#ifdef _STOUT
      call get_quarks_stout_params(this%quark,alpha,nstep)
      write(iout) this%number_of_quark,alpha,nstep
#else
      write(iout) this%number_of_quark
#endif
    else
      write(iout) this%number_of_quark
    endif
    call save_config(this%quark,iout)
  else
    write(iout) 0
  endif

  return
end subroutine

subroutine initialize_config_qcd_hmc(this,iflg,fname)
!
! Initialize/load field configuration according to HMC status
!
  use print_status_class
  use random_class
  implicit none
  type(qcd_hmc),   intent(inout) :: this
  integer,            intent(in) :: iflg
  character(CHARLEN), intent(in) :: fname


  select case (iflg)
  case (0)   ! cold start with new random number
    call set_identity(this%u)
    this%u0 = this%u
  case default
    call read_cont_config(this,fname)
    this%u0 = this%u
  end select

  return
end subroutine

subroutine rev_check_qcd_hmc(this,ifi)
!
!  MD Reversivility check routine
!
!  Save, restore, and check the differneces of configurations
!
  implicit none
  type(qcd_hmc), intent(inout) :: this   ! HMC QCD action
  integer,       intent(in)    :: ifi    ! switch, 0 for initial, 1 for final, 2 for rev check.
  type(vfield_gluon_wg),  allocatable :: duu
  type(vfield_gluon_wog), allocatable :: dpp
  real(DP) :: diff_u,diff_p

  select case(ifi)
  case(0)
  !================================
  ! Keep initial configuration
  !          (p(t=0),u(t=0))
  !================================
    if (.not. associated(m_u0)) allocate(m_u0)
    if (.not. associated(m_p0)) allocate(m_p0)
    call new(m_u0)
    call new(m_p0)
    m_u0 = this%u0
    m_p0 = this%p

  case(1)
  !================================
  ! Keep final configuration
  !         (p(t=tau),u(t=tau))
  !================================
    if (.not. associated(m_u1)) allocate(m_u1)
    if (.not. associated(m_p1)) allocate(m_p1)
    call new(m_u1)
    call new(m_p1)
    m_u1 = this%u
    m_p1 = this%p

  case(2)
  !==========================================
  ! compare inital and reversed
  !                 configuration
  ! (p(0),u(0)) vs (p(tau-tau),u(tau-tau))
  !==========================================

    !==================================================
    ! diff_u = |u(t=0) - u(t=tau-tau)|/COL/Sqrt(VOL*T)
    ! diff_p = |p(t=0) - p(t=tau-tau)|/COL/Sqrt(VOL*T)
    !==================================================
    diff_u = delta_u(m_u0,this%u)
    diff_p = delta_p(m_p0,this%p)

    if (nodeid==0) then
      write(*,'("%   DU,DP: ",2E24.16)') diff_u,diff_p
      write(*,'("% DH,DH/H: ",2E24.16)')  this%H0-this%H2 ,&
  &                                      (this%H0-this%H2)/this%H0
    endif

    !==========================
    ! restore configuration
    !         (p(tau),u(tau))
    !==========================
    this%u = m_u1
    this%p = m_p1

    if (associated(m_u0)) deallocate(m_u0)
    if (associated(m_p0)) deallocate(m_p0)
    if (associated(m_u1)) deallocate(m_u1)
    if (associated(m_p1)) deallocate(m_p1)
  end select

  return
end subroutine

subroutine measurement_onfly_qcd_hmc(this)
!
! Do measurment on the fly during HMC algorithm.
!
! This is called within the MD integrator.
!
#ifdef _SF
  use sf_measure_class
#endif
  implicit none
  type(qcd_hmc), intent(inout) :: this
  integer(8) :: mult1
  integer    :: itraj
  character(CHARLEN) :: str
#ifdef _SF
  type(sf_measurement) :: sf_obs
#endif

  itraj = get_trajectory_number(this%hmc_status)

  !------------------------
  ! save plaquette history
  !------------------------
  if (nodeid == 0) then
    mult1 = get_count(mult_iter) - this%mult0
    write(str,'(I6,10E24.16)')   &
#ifdef _RGGAUGE
 &    itraj,this%WLP1%PLQ,real(mult1,kind=DP),this%WLP1%RCT
#else
 &    itraj,this%WLP1%PLQ,real(mult1,kind=DP)
#endif
    call print(this%plaq_log,TRIM(str))
  endif

  !-------------------
  ! print status
  !-------------------
  if (nodeid == 0) then
    write(*,'("# ",I8,"  PLQ: ",F19.16,2I12)')  &
 &  itraj,this%WLP1%PLQ,get_count(mult_iter),mult1
  endif

  !-------------------------------
  ! compute reweighting factor
  !-------------------------------
  if (SW_ON == this%switch_dynamical_quark) then
    call reweighting(this%quark,this%u)
  endif

  !-------------------------------
  ! measurements
  !-------------------------------
  if (SW_ON == this%switch_measurement) then
#ifdef _SF
    call new(sf_obs,"Params_SF_PCAC",this%hmc_status)
    call measure(sf_obs,this%gluon%gluon,this%u)
    call delete(sf_obs)
#endif
    if ( .not. ASSOCIATED(this%observables) ) then
      allocate(this%observables)
      call new(this%observables,TRIM(ADJUSTL(this%measurement_fname)),this%hmc_status)
    endif
    call measure(this%observables,this%u)
  endif

  return
end subroutine

subroutine update_p_qcd_hmc(this,dt)
!
! Update Momentum (P)
!
! p = p + dt F
!
  implicit none
  type(qcd_hmc), intent(inout) :: this  ! HMC QCD action
  real(DP),      intent(in)    :: dt    ! time step for MD
  type(timer) :: etime

  call new(etime); call tic(etime)
  call force(this)

  call update_p(this%p,dt,this%F)

  if (associated(this%F)) then
    deallocate(this%F)
  endif
  call toc(etime)
  call print_log_timing(this%hmc_status,"update_p:",etime)

  return
end subroutine

subroutine update_q_qcd_hmc(this,dt)
!
! Update Corrdinate (Q)
!
! q = q + dt p
!
  implicit none
  type(qcd_hmc), intent(inout) :: this ! HMC QCD action
  real(DP),      intent(in)    :: dt   ! time step for MD
  type(timer) :: etime

  call new(etime); call tic(etime)
  call update_u(this%u,dt,this%p)
  call toc(etime)
  call print_log_timing(this%hmc_status,"update_q:",etime)

  return
end subroutine

subroutine hamil_qcd_hmc(this,ifi)
!
! Compute Total Hamiltonian for HMC QCD action
! 
  implicit none
  type(qcd_hmc), intent(inout) :: this  ! HMC QCD action
  integer,       intent(in)    :: ifi ! select target
                                      ! [0] for initial Hamiltonian with generating P.
                                      ! [1] for final Hamiltonian without generating P.
                                      ! [2] for initial Hamiltonian without generating P.

  type(tfield_gluon_wg), allocatable :: wl34,wl98
  real(DP) :: H
  real(DP) :: SU,SP
  real(DP) :: SQ_all
  real(DP) :: SQ(SIZE(this%quark))
  character(CHARLEN) :: str

  H      = 0.0_DP
  SU     = 0.0_DP
  SP     = 0.0_DP
  SQ_all = 0.0_DP
  SQ(:)  = 0.0_DP

  if (0 == ifi) then
    this%mult0 = get_count(mult_iter)
    if (.not. associated(this%p) ) then
      allocate(this%p)
      call new(this%p)
    endif
  endif

  !========================================
  ! calc. Hamiltonian for gluon action
  ! if (ifi==0)  p is updated
  !========================================
  allocate(wl34,wl98)
  call new(wl34)
  call new(wl98)
  call make_two_links(this%u,wl34,wl98)
  call hamil(this%gluon,ifi,this%u,this%p,SU,SP,wl34,wl98)

#ifndef _CLOVER
  deallocate(wl34)
  deallocate(wl98)
#endif

  !========================================
  ! calc. Hamiltonian for quark action
  !========================================
  if (SW_ON == this%switch_dynamical_quark) then
#ifdef _CLOVER
    call set_two_links(this%quark,wl34,wl98)
#endif
    call hamil(this%quark,ifi,this%u,SQ)
  endif

#ifdef _CLOVER
  call remove_two_links(this%quark)
  deallocate(wl34)
  deallocate(wl98)
#endif

  SQ_all=sum(SQ)
  H=SU+SP+SQ_all
!  Hamil%H=H
!  Hamil%SU=SU
!  Hamil%SP=SP
!  Hamil%SQ=SQ_all

  select case(ifi)
  case (0) 
    this%H0 = H
    this%WLP0 = this%gluon%WLP
  case (1)
    this%H1 = H
    this%WLP1 = this%gluon%WLP
  case (2)
    this%H2 = H
    this%WLP2 = this%gluon%WLP
  end select

  if (nodeid == 0) then
    write(str,'("@",I8,I2,"   H:",E24.16,"  SU:",E24.16,"  SP:",E24.16,"  SQ:",E24.16)') &
 &               get_trajectory_number(this%hmc_status), &
 &               ifi,H,SU,SP,SQ_all
    call print_log_action(this%hmc_status,str)
    write(str,*)
    call print_log_action(this%hmc_status,str)
  endif

  return
end subroutine

subroutine metropolis_all_qcd_hmc(this,ihit)
!
! Do all Metropolis tests for HMC QCD action
!
! This routine contains HMC Metropolis test and other Globarl Metropolis tests.
! This routine update configuration according to the result of the Metropolis test.
!
! If accepted the trial configuration U  is copied to U0,
! If rejceed  the original configuration U0 is copied to U.
!
!  After calling this Metropolis test, U is the new configuration.
! 
  implicit none
  type(qcd_hmc), intent(inout) :: this   ! HMC QCD action
  integer,       intent(out)   :: ihit   ! 1 for accepted, 0 for rejected.
  real(DP) :: dH

  if (associated(this%p)) then
    deallocate(this%p)
  endif

  dH = this%H1 - this%H0
  call metropolis_qcd_hmc(this,dH,ihit)

  if (SW_ON == this%switch_dynamical_quark) then
    call metropolis(this%quark,this%u0,this%u,ihit)
  endif

  this%mult1 = get_count(mult_iter)

  if (ihit==1) then
    this%u0   = this%u
    this%WLP0 = this%WLP1
  else
    this%u    = this%u0
    this%WLP1 = this%WLP0
  endif

  return
end subroutine

subroutine metropolis_qcd_hmc(this,dH,ihit)
!
! do Metropolis test for HMC QCD action (HMC) part.
! 
  use comlib
  use random_class
  use print_status_class
  implicit none
  type(qcd_hmc), intent(inout) :: this  ! HMC QCD action
  real(DP),      intent(in)    :: dH    ! Hamiltonian difference
  integer,       intent(out)   :: ihit  ! 1 for accepted, 0 for rejected.

  logical :: flag

!========================================
! Metropolis test (HMC),
!   P[u0 -> u1] = min[1,exp(-dH)]
! Acceptance Prob. : P[u0 -> u1],
! Energy diff.     : dH=H[u1]-H[u0]
!
! Check is done on Mastar node only.
! The master broadcasts the result.
!========================================
  flag = is_accepted(this%hmc_status,dH)

  if (flag) then
    ihit = 1
  else
    ihit = 0
  endif

  return
end subroutine

subroutine force_qcd_hmc(this)
!
! Compute MD Force with HMC QCD action
!
  use field_gauge_class, only : assign_mult, traceless_antihermite, set_sf_boundary
  implicit none
  type(qcd_hmc),  intent(inout) :: this

  type(vfield_gluon_wg), allocatable :: BB  ! pre-force
  type(tfield_gluon_wg), allocatable :: wl34,wl98
  character(CHARLEN) :: str
  real(DP) :: fnorm(2)
  type(timer) :: etime

  call new(etime);call tic(etime)
#ifdef _MD_DEPTH_
  if (nodeid==0) write(*,'("depth=",I3," : force")')this%idepth
#endif

  !=========================
  ! clear force field
  !=========================
  if (associated(this%F)) then
    deallocate(this%F)
  endif
  allocate(BB)
  call new(BB)
  call clear(BB)

  !=========================
  ! calc. 2-link field
  !=========================
  allocate(wl34,wl98)
  call new(wl34)
  call new(wl98)
  call make_two_links(this%u,wl34,wl98)
  call toc(etime)
  call print_log_timing(this%hmc_status,"make_two_links:",etime)

  !================================
  ! calc. force from gauge action
  !================================
  call force(this%gluon,BB,this%idepth,this%u,wl34,wl98)

#ifndef _CLOVER
  deallocate(wl34,wl98)
#endif

  !================================
  ! calc. force from quark action
  !================================
  if (this%switch_dynamical_quark == SW_ON) then
#ifdef _CLOVER
    call set_two_links(this%quark,wl34,wl98)
#endif
    call force(this%quark,BB,this%idepth,this%u)
#ifdef _CLOVER
    call remove_two_links(this%quark)
#endif
  endif

#ifdef _CLOVER
  deallocate(wl34,wl98)
#endif

  call tic(etime)
  !==================================================================
  !  F_mu(n) =    u_mu(n) * BB_mu(n) - (u_mu(n) * BB_mu(n))^dag
  !           -Tr[u_mu(n) * BB_mu(n) - (u_mu(n) * BB_mu(n))^dag]/3
  !==================================================================
  if (.not.associated(this%F)) then
    allocate(this%F)
    call new(this%F)
  endif
  !
  !  F = u * BB
  !
  call assign_mult(this%F,this%u,BB)
  deallocate(BB)
  !
  !  F <= F-F^+ - Tr[F-F^+]/3
  !
  call traceless_antihermite(this%F)


#ifdef _SF
  !
  ! zero clear at t=1 spatial force
  !
  call set_sf_boundary(this%F)
#endif

  call get_force_norm(fnorm,this%F)
  write(str,'(2E24.16,I3)')fnorm(1),fnorm(2),this%idepth
  call print(this%force_log,str)

  call toc(etime)
  call print_log_timing(this%hmc_status,"force_qcd:",etime)

  return
end subroutine

end module
